@@ -35,6 +35,22 @@ AWS_ECS_CLIENT = "ecs"
 AWS_CLOUDWATCH_CLIENT = "cloudwatch"
 AWS_S3_CLIENT = "s3"
 
+COL_DBNAME = "DBName"
+COL_TABLENAME = "table_name"
+COL_CREATEDATETIME = "CreateTime"
+COL_UPDATEDATETIME = "UpdateTime"
+COL_TABLETYPE = "TableType"
+COL_CREATEDBY = "CreatedBy"
+DF_TABLECOLUMNS = [
+    COL_DBNAME,
+    COL_TABLENAME,
+    COL_CREATEDATETIME,
+    COL_UPDATEDATETIME,
+    COL_TABLETYPE,
+    COL_CREATEDBY
+]
+COL_FIELDNAME = "field_name"
+COL_FIELDTYPE = "field_type"
 
 DURATION_COL_NAME = "executiontime"
 AVG_DURATION_COL_NAME = "Estimated duration (Avg mins)"
@@ -146,12 +162,14 @@ class Account(object):
         data = []
         for page in page_iterator:
             for job in page["Jobs"]:
-                data.append([job["Name"], job["CreatedOn"], job["LastModifiedOn"]])
+                data.append(
+                    [job["Name"], job["CreatedOn"], job["LastModifiedOn"]])
 
         jobs_df = pd.DataFrame(data, columns=glue_job_df_columns)
         # localize dates to remove timezone; writing to excel with timezone is not supported
         jobs_df["CreatedOn"] = jobs_df["CreatedOn"].dt.tz_localize(None)
-        jobs_df["LastModifiedOn"] = jobs_df["LastModifiedOn"].dt.tz_localize(None)
+        jobs_df["LastModifiedOn"] = jobs_df["LastModifiedOn"].dt.tz_localize(
+            None)
 
         return jobs_df
 
@@ -283,7 +301,7 @@ class Account(object):
         """
         dbs = wr.catalog.get_databases(boto3_session=self.session)
         data = [(db["Name"]) for db in dbs]
-        databases_df = pd.DataFrame(data, columns=["DBName"])
+        databases_df = pd.DataFrame(data, columns=[COL_DBNAME])
         return databases_df
 
     def glue_get_tables(self, dbname=None) -> pd.DataFrame:
@@ -291,12 +309,13 @@ class Account(object):
         Get AWS Glue tables
 
         Keyword Arguments:
-            dbname -- Database Name for which to retrive tables (default: {None})
+            dbname -- Database Name for which to retrieve tables (default: {None})
 
         Returns:
             dataframe
         """
-        tables = wr.catalog.get_tables(database=dbname, boto3_session=self.session)
+        tables = wr.catalog.get_tables(
+            database=dbname, boto3_session=self.session)
 
         data = [
             (
@@ -309,19 +328,43 @@ class Account(object):
             for table in tables
         ]
 
-        # TODO: fix hardcodings in next line
         tables_df = pd.DataFrame(
             data,
             columns=[
-                "table_name",
-                "CreateTime",
-                "UpdateTime",
-                "TableType",
-                "CreatedBy",
+                COL_TABLENAME,
+                COL_CREATEDATETIME,
+                COL_UPDATEDATETIME,
+                COL_TABLETYPE,
+                COL_CREATEDBY,
             ],
         )
         return tables_df
 
+    def glue_get_fields(self, dbname, tablename) -> pd.DataFrame:
+        """
+        Get AWS Glue table columns
+
+        Keyword Arguments:
+            dbname -- Database Name for table 
+            tablename -- Database Name for which to retrieve columns 
+
+        Returns:
+            dataframe
+        """
+        dict_columns = wr.catalog.get_table_types(
+            database=dbname, table=tablename, boto3_session=self.session)
+
+        df_data = []
+        for key in dict_columns:
+            df_data.append([tablename, key, dict_columns[key]])
+
+        column_names = [COL_TABLENAME,
+                        COL_FIELDNAME,
+                        COL_FIELDTYPE]
+        df_columns = pd.DataFrame(columns=column_names, data=df_data)
+        # print(df_columns.head())
+        return df_columns
+
     def athena_execute_query(self, database: str, query: str, use_cache: bool = True):
         max_cache_seconds = 172800
         if not use_cache:
@@ -338,7 +381,8 @@ class Account(object):
             )
         except exceptions.QueryFailed as error:
             # TODO: log Query execution error
-            print(f"Error executing query for table: {exceptions.QueryFailed.__name__}")
+            print(
+                f"Error executing query for table: {exceptions.QueryFailed.__name__}")
             raise error
         except botocore.exceptions.ClientError as error:
             # TODO: log Query execution error
@@ -373,7 +417,8 @@ class Account(object):
         executionid = response["QueryExecutionId"]
         logger.debug(executionid)
 
-        response = client_athena.get_query_execution(QueryExecutionId=executionid)
+        response = client_athena.get_query_execution(
+            QueryExecutionId=executionid)
         OutputLocation = response["QueryExecution"]["ResultConfiguration"][
             "OutputLocation"
         ]
@@ -417,6 +462,73 @@ class Account(object):
 
         logger.debug("Athena table created successfully.")
 
+    def athena_data_dictionary(self, include_dbs: list = [], exclude_dbs: list = []) -> pd.DataFrame:
+        """
+        Get AWS Athean data dictionary. A data frame with all databases, tables & fields with datatypes
+
+        Keyword Arguments:
+            include_dbs (optional) -- list of databases to be included
+            exclude_dbs (optional) -- list of databases to be excluded if include_dbs list is empty.
+
+        Returns:
+            dataframe
+        """
+        # Get glue databases
+        databases_df = self.glue_get_databases()
+
+        # create list of databases and loop to retrieve all fields
+        lst_databases = databases_df[COL_DBNAME].tolist()
+
+        final_tables_df = pd.DataFrame(columns=DF_TABLECOLUMNS)
+        final_fields_df = pd.DataFrame(
+            columns=[COL_DBNAME, COL_TABLENAME, COL_FIELDNAME, COL_FIELDTYPE])
+        lst_tqdm = tqdm(lst_databases)
+        for db in lst_tqdm:
+            lst_tqdm.set_description(f"Processing database: {db}")
+            if len(include_dbs) > 0:
+                if not db in include_dbs:
+                    continue
+            elif len(include_dbs) == 0 and len(exclude_dbs) > 0:
+                if db in exclude_dbs:  # These databases are not required
+                    continue
+
+            tables_df = self.glue_get_tables(dbname=db)
+            tables_df.insert(0, COL_DBNAME, db, allow_duplicates=True)
+
+            # Get fields for tables
+            lst_tables = tables_df[COL_TABLENAME].tolist()
+            for table in lst_tables:
+                fields_df = self.glue_get_fields(dbname=db, tablename=table)
+                fields_df.insert(0, COL_DBNAME, db, allow_duplicates=True)
+                # print(fields_df.head())
+
+                final_fields_df = pd.concat([final_fields_df, fields_df])
+
+            final_tables_df = pd.concat([final_tables_df, tables_df])
+
+        # Merge tables & fields
+        print(f"Shape before fields merge: {final_tables_df.shape}")
+        final_tablesAndFields_df = pd.merge(
+            final_tables_df, final_fields_df, how="left", on=[COL_DBNAME, COL_TABLENAME])
+        print(f"Shape after fields merge: {final_tables_df.shape}")
+
+        # TODO: Check duplicates here
+        # duplicate_count = final_df.duplicated(subset=[COL_DBNAME, COL_TABLENAME]).sum()
+        # if duplicate_count > 0:
+        #     print(f"Duplicate tables found. Duplicate count = {duplicate_count}")
+        #     exit(0)
+        # else:
+        #     print(f"No duplicate tables.")
+
+        # localize dates to remove timezone; writing to excel with timezone is not supported
+        # final_tablesAndFields_df[COL_CREATEDATETIME] = final_tablesAndFields_df[COL_CREATEDATETIME].dt.tz_localize(
+        #     None)
+        # final_tablesAndFields_df[COL_UPDATEDATETIME] = final_tablesAndFields_df[COL_UPDATEDATETIME].dt.tz_localize(
+        #     None)
+
+        # final_df.to_csv("dl_datadictionary.csv", index=False)
+        return final_tablesAndFields_df
+
     def lambda_get_functions(self):
         dataframe_columns = [
             "arn",
@@ -454,7 +566,8 @@ class Account(object):
     def lambda_get_metrics_list(self, namespace="AWS/Lambda"):
         # TODO: this is incomplete
         logger.debug("Retrieving lambda metrics list")
-        client_cloudwatch: boto3.client = self.session.client(AWS_CLOUDWATCH_CLIENT)
+        client_cloudwatch: boto3.client = self.session.client(
+            AWS_CLOUDWATCH_CLIENT)
         paginator = client_cloudwatch.get_paginator("list_metrics")
         page_iterator = paginator.paginate(Namespace=namespace)
 
@@ -489,7 +602,8 @@ class Account(object):
         edate = datetime.now() if end_date is None else end_date
 
         logger.debug(f"Retrieving invocations for lambda={lambda_name}")
-        client_cloudwatch: boto3.client = self.session.client(AWS_CLOUDWATCH_CLIENT)
+        client_cloudwatch: boto3.client = self.session.client(
+            AWS_CLOUDWATCH_CLIENT)
         paginator = client_cloudwatch.get_paginator("get_metric_data")
         page_iterator = paginator.paginate(
             MetricDataQueries=[
@@ -672,7 +786,8 @@ class Account(object):
 
         all_services_df = pd.DataFrame(columns=["clusterArn", "serviceArn"])
         tqdm_cluster = tqdm(lst_clusers)
-        tqdm_cluster.set_description("Gathering ecs clusters & services metadata")
+        tqdm_cluster.set_description(
+            "Gathering ecs clusters & services metadata")
         for cluster in tqdm_cluster:
             services_df = self.ecs_get_services(cluster_arn=cluster)
             all_services_df = pd.concat([all_services_df, services_df])
@@ -745,7 +860,8 @@ class Account(object):
         for page in page_iterator:
             for reservation in page["Reservations"]:
                 for instance in reservation["Instances"]:
-                    tag_name = [tag for tag in instance["Tags"] if tag["Key"] == "Name"]
+                    tag_name = [tag for tag in instance["Tags"]
+                                if tag["Key"] == "Name"]
                     row = (
                         instance["InstanceId"],
                         tag_name[0]["Value"],
